{-# LANGUAGE Arrows, CPP #-}
module Controller where

#if __GLASGOW_HASKELL__ >= 610
import Prelude hiding ((.), id)
import Control.Category
#endif
import Control.Arrow
import Control.Monad (unless)

import qualified Data.IntMap as IM
import Data.Complex
import Data.IORef
import Data.Either (either)
import Text.Printf

import Phisics
import OrbitVM
import SubmissionFile

-- -----------------------------------------------------------

newtype Automata i o = Automata{ runAutomata :: i -> (o, Automata i o) }

#if __GLASGOW_HASKELL__ >= 610
instance Category Automata where
    id = Automata $ \x -> (x, id)
    f . g  = Automata $ \x ->
      let (y, g') = runAutomata g x
          (z, f') = runAutomata f y
      in (z, f' . g')

instance Arrow Automata where
    arr f = let a = Automata $ \x -> (f x, a) in a
    first f = Automata $ \(x, y) ->
      let (z, f') = runAutomata f x
      in ((z, y), first f')
#elif
instance Arrow Automata where
    g >>> f = Automata $ \x ->
      let (y, g') = runAutomata g x
          (z, f') = runAutomata f y
      in (z, g' >>> f')
    arr f = let a = Automata $ \x -> (f x, a) in a
    first f = Automata $ \(x, y) ->
      let (z, f') = runAutomata f x
      in ((z, y), first f')
#endif

instance ArrowLoop Automata where
    loop f = Automata $ \x ->
      let ((z,y), f') = runAutomata f (x,y)
      in (z, loop f')

constA :: b -> Automata a b
constA b = arr (\_ -> b)

fby :: a -> Automata a a
fby x = Automata $ \y -> (x, fby y)

diff :: Num b => Automata a b -> Automata a (Maybe b)
diff a = proc x -> do
  y  <- a -< x
  y' <- fby Nothing <<< arr Just -< y
  returnA -< fmap (y -) y'

-- -----------------------------------------------------------

type Input = (Time, SensorData)
type Controller = Automata Input Acc

-- 現在時刻
time :: Automata Input Int
time = arr fst

-- センサーの値
sensor :: Automata Input SensorData
sensor = arr snd

-- スコア
score :: Automata Input Double
score = sensor >>> arr (IM.! addrScore)

-- 自衛星の位置 (地球原点)
pos :: Automata Input Pos
pos = proc i -> do
  s <- sensor -< i
  returnA -< (- s IM.! addrSx) :+ (- s IM.! addrSy)

-- 自衛星の速度
velocity :: Automata Input (Maybe Vel)
velocity = diff pos

-- 自衛星の残燃料
fuelRemaining :: Automata Input Double
fuelRemaining = sensor >>> arr (IM.! addrFuelRemaining)

-- 目標とする半径
targetRadius :: Automata Input Double
targetRadius = sensor >>> arr (IM.! addrTargetRadius)

-- ターゲットの衛星の位置 (地球原点)
targetSatelitePos :: Automata Input Pos
targetSatelitePos = proc i -> do
  s <- sensor -< i
  p <- pos -< i
  let q = (s IM.! 0x4) :+ (s IM.! 0x5) -- relative to target satellite
  returnA -< p + q

-- ターゲットの衛星の速度
targetSateliteVelocity :: Automata Input (Maybe Vel)
targetSateliteVelocity = diff targetSatelitePos

-- -----------------------------------------------------------

newtype Task a b c = Task{ runTask :: Automata a (Either b c) }

instance Monad (Task a b) where
  return x = Task $ arr (\_ -> Right x)
  -- 綺麗に書けなかった
  m >>= f = Task $ Automata $ \a ->
    case runAutomata (runTask m) a of
      (Left b, am) -> (Left b, runTask (Task am >>= f))
      (Right c, _) -> runAutomata (runTask (f c)) a 

mkTask :: Automata a (b, Maybe c) -> Task a b c
mkTask a = Task (a >>> arr f)
  where
    f :: (b, Maybe c) -> Either b c
    f (b, Nothing) = Left b
    f (b, Just c)  = Right c

runTask_ :: Task a b c -> Automata a b
runTask_ t = runTask t >>> arr (either id (const undefined))

constT :: b -> Task a b c
constT b = Task $ arr (\_ -> Left b)

snapT :: Task a b a
snapT = Task $ arr (\a -> Right a)

dmapTask :: (a -> a') -> (b' -> b) -> Task a' b' c -> Task a b c
dmapTask f g t = Task $ arr f >>> runTask t >>> arr h
  where
    h (Left b') = Left (g b')
    h (Right c) = Right c

{-
timeOut :: Task a b c -> Int -> Task a b (Maybe c)
abortWhen :: Task a b c -> Automata a (Maybe d) -> Task a b (Either c d)
-}

output :: b -> Task a b ()
output b = Task $ Automata $ \_ -> (Left b, constA (Right ()))

-- -----------------------------------------------------------

wait :: Int -> Task a Acc ()
wait n = sequence_ $ replicate n (output 0)

hohmannTransfer :: Vel -> Double -> Double -> Task a Acc ()
hohmannTransfer v r1 r2 = do
  output (mkPolar dv1 theta1)
  wait t_h
  output (mkPolar dv2 theta2)
  return ()
  where
    theta1 = phase v
    theta2 = theta1 + pi
    dv1 = sqrt (mu / r1) * (sqrt (2 * r2 / (r1+r2)) - 1)
    dv2 = sqrt (mu / r2) * (1 - sqrt (2 * r1 / (r1+r2)))
    t_h = round (pi * sqrt ((r1+r2)**3 / (8*mu)))

-- -----------------------------------------------------------

control :: IOVM -> Command -> Controller -> IO [Frame]
control = controlWithTrace (\_ _ -> return ())

controlWithTrace :: (Time -> SensorData -> IO ()) -> IOVM -> Command -> Controller -> IO [Frame]
controlWithTrace trace vm init controller = do
  r <- newIORef newRecorder
  let step x = do
        t <- vmtime vm
        recorder <- readIORef r
        writeIORef r $! record recorder t x
        let x' = IM.filter (/= 0) x
        unless (IM.null x') $ printf "vm(t=%d): %s\n" t (show (IM.toList x'))
        oneStepM vm x
  o0 <- step init
  let go c o = do
        t <- vmtime vm
        trace t o
        let score = o IM.! addrScore
        if (score == 0)
          then do
            let (a, c') = runAutomata c (t, o)
            o' <- step (IM.fromList [(addrDVx, realPart a), (addrDVy, imagPart a)])
            go c' o'
          else do
            recorder <- readIORef r
            putStrLn $ "Score: " ++ show score
            return $ recordFin recorder t
  go controller o0
