{-# LANGUAGE BangPatterns, FlexibleContexts, ScopedTypeVariables, MultiParamTypeClasses #-}
{-# OPTIONS_GHC -Wall #-}
module OrbitVM where

import Control.Monad
import Control.Monad.Trans
import Control.Monad.State
import Data.Array.MArray
import qualified Data.IntMap as IM
import Data.Word
import Data.Maybe (fromMaybe)

import Control.Monad.ST
import Data.Array.ST
import Data.STRef

import Data.Array.IO
import Data.IORef

import System.IO
import Foreign.Marshal (alloca)
import Foreign.Storable (peek)
import Data.Bits

import Text.Printf

type Data = Double
type Addr = Word16

addrMax :: Int
addrMax = (1 `shiftL` 14) - 1

addrConf, addrDVx, addrDVy :: Int
addrConf = 0x3e80
addrDVx = 0x2
addrDVy = 0x3

addrScore, addrFuelRemaining, addrSx, addrSy, addrTargetRadius :: Int
addrScore = 0x0
addrFuelRemaining = 0x1
addrSx = 0x2
addrSy = 0x3
addrTargetRadius = 0x4

data Inst
    = Add    {-# UNPACK #-} !Addr {-# UNPACK #-} !Addr
    | Sub    {-# UNPACK #-} !Addr {-# UNPACK #-} !Addr
    | Mult   {-# UNPACK #-} !Addr {-# UNPACK #-} !Addr
    | Div    {-# UNPACK #-} !Addr {-# UNPACK #-} !Addr
    | Output {-# UNPACK #-} !Addr {-# UNPACK #-} !Addr
    | Phi    {-# UNPACK #-} !Addr {-# UNPACK #-} !Addr
    | Noop
    | Cmpz !CmpOp {-# UNPACK #-} !Addr
    | Sqrt  {-# UNPACK #-} !Addr
    | Copy  {-# UNPACK #-} !Addr
    | Input {-# UNPACK #-} !Addr
    deriving (Show)

data CmpOp
    = LTZ
    | LEZ
    | EQZ
    | GEZ
    | GTZ
    deriving (Show, Eq, Ord)

evalCmpOp :: CmpOp -> (Double -> Double -> Bool)
evalCmpOp LTZ = (<)
evalCmpOp LEZ = (<=)
evalCmpOp EQZ = (==)
evalCmpOp GEZ = (>=)
evalCmpOp GTZ = (>)

class Monad m => Ref r m where
  readRef  :: r a -> m a
  writeRef :: r a -> a -> m ()
  modifyRef :: r a -> (a -> a) -> m ()
  modifyRef r f = readRef r >>= \x -> writeRef r (f x)

instance Ref (STRef s) (ST s) where
  readRef  = readSTRef
  writeRef = writeSTRef
  modifyRef = modifySTRef

instance Ref IORef IO where
  readRef  = readIORef
  writeRef = writeIORef
  modifyRef = modifyIORef

type VM a r = ([Inst], a Addr Data, r Bool, r Int)
type STVM s = VM (STUArray s) (STRef s)
type IOVM = VM IOUArray IORef

type Time = Int

vmtime :: Ref r m => VM a r -> m Time
vmtime (_, _, _, time) = readRef time

type Command = IM.IntMap Data
type SensorData = IM.IntMap Data

{-# SPECIALIZE oneStepM :: STVM s -> Command -> ST s SensorData #-}
{-# SPECIALIZE oneStepM :: IOVM -> Command -> IO SensorData #-}
oneStepM
  :: forall m a r. (MArray a Data m, Ref r m)
  => VM a r        -- instructions, memory and status
  -> Command       -- inport
  -> m SensorData  -- outport
oneStepM (inst, mem, status, time) inport = modifyRef time (+1) >> execStateT (go 0 inst) IM.empty
  where
    go :: Addr -> [Inst] -> StateT SensorData m ()
    go pc [] = seq pc $ return ()
    go pc (i:is) = seq pc $ do
      case i of
         Add  r1 r2 -> lift $ liftM2 (+) (readArray mem r1) (readArray mem r2) >>= writeArray mem pc
         Sub  r1 r2 -> lift $ liftM2 (-) (readArray mem r1) (readArray mem r2) >>= writeArray mem pc
         Mult r1 r2 -> lift $ liftM2 (*) (readArray mem r1) (readArray mem r2) >>= writeArray mem pc
         Div  r1 r2 -> lift $ liftM2 quo (readArray mem r1) (readArray mem r2) >>= writeArray mem pc
           where
             quo x y = if y==0 then 0 else x / y             
         Output r1 r2 -> do
             x <- lift $ readArray mem r2
             modify $ IM.insert (fromIntegral r1) x
         Phi r1 r2 -> lift $ do
             s <- readRef status
             readArray mem (if s then r1 else r2) >>= writeArray mem pc
         Noop -> return ()
         Cmpz op r1 -> lift $ do
             x <- readArray mem r1
             writeRef status $! evalCmpOp op x 0.0
         Sqrt  r1 -> lift $ liftM sqrt (readArray mem r1) >>= writeArray mem pc
         Copy  r1 -> lift $ readArray mem r1 >>= writeArray mem pc
         Input r1 -> lift $ writeArray mem pc $ fromMaybe 0 (IM.lookup (fromIntegral r1) inport)
      go (pc+1) is

newMem :: MArray a Data m => [Data] -> m (a Addr Data)
newMem ds = do
  mem <- newArray (0, fromIntegral addrMax) 0
  sequence_ [ writeArray mem i d | (i, d) <- zip [0..] ds ]
  return mem

load :: FilePath -> IO ([Inst], [Data])
load fname = do
  h <- openBinaryFile fname ReadMode
  alloca $ \pdata -> alloca $ \pinst -> do
    let go evenframe is ds = seq evenframe $ do
          eof <- hIsEOF h
          if eof
            then return (reverse is, reverse ds)
            else do
              if evenframe
                then do
                  hGetBuf h pdata 8
                  hGetBuf h pinst 4
                else do
                  hGetBuf h pinst 4
                  hGetBuf h pdata 8
              d <- peek pdata
              i <- liftM decodeInst $ peek pinst
              go (not evenframe) (i : is) (d : ds)
    go True [] []

decodeInst :: Word32 -> Inst
decodeInst x =
    case d_op of
      1 -> Add  d_r1 d_r2
      2 -> Sub  d_r1 d_r2
      3 -> Mult d_r1 d_r2
      4 -> Div  d_r1 d_r2
      5 -> Output d_r1 d_r2
      6 -> Phi d_r1 d_r2
      0 ->
        case s_op of
          0 -> Noop
          1 -> Cmpz op s_r1
            where
              op = case s_imm `shiftR` 7 of
                     0 -> LTZ
                     1 -> LEZ
                     2 -> EQZ
                     3 -> GEZ
                     4 -> GTZ
                     _ -> error "unknown comparison operator"
          2 -> Sqrt s_r1
          3 -> Copy s_r1
          4 -> Input s_r1
          _ -> error "unknown S-type instruction"
      _ -> error "unknown D-type instruction"
  where
    mask14 = (1 `shiftL` 14) - 1
    mask10 = (1 `shiftL` 10) - 1
    d_op = x `shiftR` 28
    d_r1 = fromIntegral $ (x `shiftR` 14) .&. mask14
    d_r2 = fromIntegral $ x .&. mask14
    s_op = x `shiftR` 24
    s_imm = (x `shiftR` 14) .&. mask10
    s_r1  = fromIntegral $ x .&. mask14

newIOVM :: FilePath -> IO IOVM
newIOVM fname = do
  (is, ds) <- load fname
  mem <- newMem ds
  status <- newIORef False
  time <- newIORef 0
  return (is, mem, status, time)
