{-# LANGUAGE Arrows #-}
module SDLControl (startViewer) where

import OrbitVM
import Controller

import Control.Monad
import qualified Data.IntMap as IM
import Data.Complex

import Multimedia.SDL
import System.Time
import Data.IORef
import Control.Concurrent
import Text.Printf
import System.IO

import Debug.Trace

wndTitle = "Orbit"
wndSize = sz 640 480

elapseTime :: Integer -> IO (IO (Int,Bool))
elapseTime fps = do
  let frametime = picosec `div` fps
  tm <- getClockTime
  st <- newIORef ((0,0,noTimeDiff),(1,tm))
  return $ do
    ((bef,cur,fdt),(cnt,bt)) <- readIORef st
    ct       <- getClockTime
    let dt   = diffClockTimes ct bt
        ndt  = diffClockTimes ct tm
        adj  = frametime*cnt - toPsec dt
        nc   = if cnt==fps then (1,ct) else (cnt+1,bt)
        (nbef,ncur) = if tdSec fdt /= tdSec ndt then (cur,0) else (bef,cur)
    if adj<0 then do
        writeIORef st ((nbef,ncur,ndt),nc)
        return (bef,False)
      else do
        writeIORef st ((nbef,ncur+1,ndt),nc)
        threadDelay $ fromInteger $ min 16666 $ adj `div` 1000000
        return (bef,True)
  where
    toPsec dt = toInteger (tdMin dt * 60 + tdSec dt) * picosec + tdPicosec dt
    picosec = 1000000000000

loop et od bef = do
  quit <- checkEvent
  when (not quit) $ do
    ks <- getKeyState
    (fps,draw) <- et
    when draw $ od
    loop et od ks
  where
    toPsec dt = toInteger (tdMin dt * 60 + tdSec dt) * picosec + tdPicosec dt
    picosec = 1000000000000

checkEvent = do
  ev <- pollEvent
  case ev of
    Just QuitEvent -> return True
    Just (KeyboardEvent { kbPress = True, kbKeysym = Keysym { ksSym = ks, ksMod = km } })
      | ks == SDLK_ESCAPE -> return True
      | ks == SDLK_F4 && (KMOD_LALT `elem` km ||
                          KMOD_RALT `elem` km) -> return True
    Nothing        -> return False
    _              -> checkEvent

sdlStart fs title (Size w h) p = do
  True <- sdlInit fs
  setCaption title title
  sur  <- setVideoMode w h 32 [HWSURFACE,DOUBLEBUF,ANYFORMAT]
  p sur
  sdlQuit

onDraw :: Int -> Double -> Surface -> IORef (Maybe SensorData) -> IO ()
onDraw problem scale sur ref = do
  fillRect sur Nothing 0x000000
  o <- readIORef ref
  case o of
    Nothing -> return ()
    Just o -> do
        let ox = 320
            oy = 240
        fillRect sur (Just $ Rect (ox-5) (oy-5) 10 10) 0x0000FF
        let pos = (- (o IM.! addrSx)) :+ (- (o IM.! addrSy)) -- 地球からの相対座標
            x = ox + round (realPart pos / scale)
            y = oy - round (imagPart pos / scale)
        fillRect sur (Just $ Rect (x-10) (y-10) 20 20) 0xFFFFFF
        when (problem == 2 || problem == 3) $ do
          let pos' = pos + ((o IM.! 0x4) :+ (o IM.! 0x5))
              x = ox + round (realPart pos' / scale)
              y = oy - round (imagPart pos' / scale)
          fillRect sur (Just $ Rect (x-10) (y-10) 20 20) 0xFF0000
          return ()
        when (problem == 4) $ flip catch (\e -> print e) $ do
          -- fuel station
          let pos' = pos + ((o IM.! 0x4) :+ (o IM.! 0x5))
              x = ox + round (realPart pos' / scale)
              y = oy - round (imagPart pos' / scale)
          fillRect sur (Just $ Rect (x-10) (y-10) 20 20) 0x00FF00
          forM_ [0..10] $ \k -> do
            let pos' = pos + ((o IM.! (3*k + 0x7)) :+ (o IM.! (3*k + 0x8)))
                x = ox + round (realPart pos' / scale)
                y = oy - round (imagPart pos' / scale)
            fillRect sur (Just $ Rect (x-10) (y-10) 20 20) 0xFF0000 >> return ()
        return ()
  flipSurface sur
  return ()

startViewer :: Int -> Double -> IOVM -> IORef (Maybe SensorData) -> IO ()
startViewer problem scale vm ref = sdlStart [VIDEO] wndTitle wndSize  $ \sur -> do
  et <- elapseTime 60
  loop et (onDraw problem scale sur ref) []
